"""Build c_resolve cython extension."""


def build(setup_kwargs):
    """Needed for the poetry building interface."""

    try:
        from Cython.Build import cythonize
    except ImportError:
        pass
    else:
        # use cythonize to build the extensions
        modules = [
            "src/buvar/di/c_resolve.pyx",
        ]

        extensions = cythonize(modules)
        setup_kwargs.update({
            'ext_modules': extensions,
            'include_dirs': [],
        })



import logging
import logging.config
import os
import sys

import orjson
import structlog

LOGGING_LEVEL_NAMES = list(map(logging.getLevelName, sorted((
    logging.NOTSET, logging.DEBUG, logging.INFO,
    logging.WARN, logging.ERROR, logging.CRITICAL,
))))
DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.WARNING)
PID = os.getpid()


class ExtractLogExtra:      # noqa: R0903

    """Extract log record attributes to structlog event_dict."""

    def __init__(self, *attrs):
        self.attrs = attrs

    def __call__(self, logger, method_name, event_dict):
        """
        Add the logger name to the event dict.
        """
        record = event_dict.get("_record")
        for attr_name in self.attrs:
            if hasattr(record, attr_name):
                attr = getattr(record, attr_name)
                event_dict[attr_name] = attr
        return event_dict


class StdioToLog:

    """Delegate sys.stdout to a logger."""

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):   # noqa
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):        # noqa
        pass


def add_os_pid(logger, method_name, event_dict):        # noqa
    """Add the logger name to the event dict."""
    event_dict["pid"] = PID
    return event_dict


class OrJsonRenderer:
    def __init__(self, option=orjson.OPT_NAIVE_UTC):    # noqa: I1101
        self.option = option

    def __call__(self, logger, name, event_dict):
        return orjson.dumps(event_dict,                 # noqa: I1101
                            default=lambda obj: str(repr(obj)),
                            option=self.option)


def setup_logging(tty=True, level=logging.DEBUG, capture_warnings=True,
                  redirect_print=False):
    """Set up structured logging for logstash.

    :param tty: if True renders colored logs
    :param level: the log level to be applied
    :param capture_warnings: redirect warnings to the logger
    :param redirect_print: use the logger to redirect printed messages
    """
    # normalize level
    if isinstance(level, str):
        level = logging.getLevelName(level.upper())

    renderer = (structlog.dev.ConsoleRenderer()
                if tty else OrJsonRenderer()
                )
    timestamper = structlog.processors.TimeStamper(fmt="ISO", utc=True)
    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        add_os_pid,
        structlog.processors.format_exc_info,
        ExtractLogExtra('spec', 'url', 'mimetype', 'has_body', 'swagger_yaml',
                        'method', 'path', 'operation_id', 'data'),
        timestamper,
    ]

    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": renderer,
                "foreign_pre_chain": pre_chain,
            },
            "colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": renderer,
                "foreign_pre_chain": pre_chain,
            },
        },
        "handlers": {
            "default": {
                # "level": "DEBUG",
                "class": "logging.StreamHandler",
                "formatter": "colored",
            },
            # "file": {
            #     # "level": "DEBUG",
            #     "class": "logging.handlers.WatchedFileHandler",
            #     "filename": "test.log",
            #     "formatter": "plain",
            # },
        },
        "loggers": {
            "": {
                "handlers": ["default"],
                "level": level,
                "propagate": True,
            },
        }
    })
    logging.captureWarnings(capture_warnings)
    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        add_os_pid,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    structlog.configure(
        processors=processors,
        context_class=structlog.threadlocal.wrap_dict(dict),
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    if redirect_print:
        # redirect stdio print
        print_log = structlog.get_logger('print')
        sys.stderr = StdioToLog(print_log)
        sys.stdout = StdioToLog(print_log)

    # log uncaught exceptions
    sys.excepthook = uncaught_exception


def uncaught_exception(ex_type, ex_value, tb):  # noqa: C0103
    log_ = structlog.get_logger('sys.excepthook')
    log_.critical(event='uncaught exception', exc_info=(ex_type, ex_value, tb))



import os
import sys
import typing

import attr
import structlog
import tomlkit

import cattr

CNF_KEY = "buvar_config"


logger = structlog.get_logger()


missing = object()


@attr.s(auto_attribs=True)
class ConfigValue:
    name: str = None
    help: str = None


def var(
    default=missing,
    converter=None,
    factory=missing,
    name=None,
    validator=None,
    help=None,
):  # noqa: W0622
    return attr.ib(
        metadata={CNF_KEY: ConfigValue(name, help)},
        converter=converter,
        validator=validator,
        **({"default": default} if factory is missing else {"factory": factory}),
    )


def _env_to_bool(val):
    """
    Convert *val* to a bool if it's not a bool in the first place.
    """
    if isinstance(val, bool):
        return val
    val = val.strip().lower()
    if val in ("1", "true", "yes", "on"):
        return True

    return False


def _env_to_list(val):
    """Take a comma separated string and split it."""
    if isinstance(val, str):
        val = map(lambda x: x.trim(), val.split(","))
    return val


def bool_var(default=missing, name=None, help=None):  # noqa: W0622
    return var(default=default, name=name, converter=_env_to_bool, help=help)


def list_var(default=missing, name=None, help=None):  # noqa: W0622
    return var(default=default, name=name, converter=_env_to_list, help=help)


@attr.s(auto_attribs=True)
class BuvarConfig:
    log_level: str = var(help="The log level to set")
    include: typing.List[str] = list_var(help="A list of plugins to include")


class ConfigSource(dict):

    """Config dict, with loadable sections.

    >>> @attr.s(auto_attribs=True)
    ... class FooConfig:
    ...     bar: float
    >>>
    >>>
    >>> source = {'foo': {'bar': 1.23}}
    >>> config = ConfigSource(source)
    >>> foo = config.load(FooConfig, "foo")
    >>> foo
    FooConfig(bar=1.23)
    """

    __slots__ = ("env_prefix",)

    def __init__(self, *sources, env_prefix=None):
        super().__init__()
        for source in sources:
            merge_dict(self, source)
        # config = schematize(__source, cls, env_prefix=env_prefix)
        self.env_prefix = env_prefix  # noqa: W0212

    def merge(self, source):
        merge_dict(self, source)

    def load(self, scheme, name=...):
        if name is ...:
            values = self
        else:
            values = self.get(name, {})

        scheme = schematize(
            scheme,
            values,
            env_prefix="_".join(
                part
                for part in (self.env_prefix, name.upper() if name is not ... else None)
                if part
            ),
        )
        return scheme


def isattrs(obj):
    return hasattr(obj, "__attrs_attrs__")


def merge_dict(dest, source):
    """Merge `source` into `dest`.

    `dest` is altered in place.
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = dest.setdefault(key, {})
            merge_dict(node, value)
        else:
            dest[key] = value
    return dest


def isunion(hint):
    origin = getattr(hint, "__origin__", None)
    return origin is typing.Union


def isoptional(hint):
    return isunion(hint) and len(hint.__args__) == 2 and type(None) in hint.__args__


def optional_type(hint):
    return hint.__args__[0]


def schematize(attrs, source, env_prefix=""):
    """Create a scheme from the source dict and apply environment vars."""
    attrs_kwargs = {}
    assert isattrs(attrs), f"{attrs} must be attrs decorated"
    for attrib in attrs.__attrs_attrs__:
        hints = typing.get_type_hints(attrs)
        a_type = attrib.type
        a_name = attrib.name
        a_hint = hints[a_name]
        env_name = "_".join(part for part in (env_prefix, a_name.upper()) if part)
        try:
            if isattrs(a_type):
                a_value = schematize(a_type, source[a_name], env_prefix=env_name)
            else:
                a_value = os.environ.get(env_name, source.get(a_name, missing))

                if a_value is missing:
                    if isoptional(a_hint):
                        # a_type = optional_type(a_type)
                        a_value = None
                    else:
                        if attrib.default is missing:
                            raise ValueError("Attribute is missing", a_name, env_name)
                        # we skip this value if source is lacking but we have a default
                        continue
                elif attrib.converter is None:
                    a_value = cattr.structure(a_value, a_type)

            attrs_kwargs[a_name] = a_value
        except KeyError:
            logger.warn("Key not in source", key=a_name, source=source)

    scheme = attrs(**attrs_kwargs)
    return scheme


def generate_env_help(attrs, env_prefix=""):
    """Generate a list of all environment options."""

    help = "\n".join(  # noqa: W0622
        "_".join((env_prefix,) + path if env_prefix else path).upper()
        for path, attrib in traverse_attrs(attrs)
        if not isattrs(attrib)
    )
    return help


def generate_toml_help(attrs, env_prefix="", parent=None):
    if parent is None:
        parent = tomlkit.table()
        doclines = trim(attrs.__doc__).split("\n")
        for line in doclines:
            parent.add(tomlkit.comment(line))
        parent.add(tomlkit.nl())

    for attrib in attrs.__attrs_attrs__:
        meta = attrib.metadata.get(CNF_KEY)
        if isattrs(attrib.type):
            # yield (attrib.name,), attrib
            sub_doc = generate_toml_help(attrib.type)
            parent.add(attrib.name, sub_doc)
        else:
            if meta:
                parent.add(tomlkit.comment(meta.help))

            if attrib.default not in (missing, attr.NOTHING):
                default = (
                    attrib.default() if callable(attrib.default) else attrib.default
                )

                parent.add(attrib.name, default)
            else:
                parent.add(tomlkit.comment(f"{attrib.name} ="))

            parent.add(tomlkit.nl())

    return parent


def traverse_attrs(attrs, with_nodes=False):
    for attrib in attrs.__attrs_attrs__:
        if isattrs(attrib.type):
            if with_nodes:
                yield (attrib.name,), attrib
            for path, sub_attrib in traverse_attrs(attrib.type):
                yield (attrib.name,) + path, sub_attrib
        else:
            yield (attrib.name,), attrib


def trim(docstring):
    # https://www.python.org/dev/peps/pep-0257/
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return "\n".join(trimmed)



import collections
import inspect
import sys
import typing

import toml

from buvar import components, config


class CliArgs(list):
    def __init__(self, args=sys.argv[1:]):
        self.args = list(args)

    def take(self, i):
        try:
            return self.args[:i]
        finally:
            self.args = self.args[i:]


class Parse:
    def __init__(self):
        # holds all
        self.stack: typing.Dict[
            typing.Callable,
            typing.List[Argument]
        ] = collections.OrderedDict()

    def argument(self, *args, **kwargs):
        argument = (
            args[0](self, *args[1:], **kwargs)
            if inspect.isclass(args[0])
            else Argument(self, *args, **kwargs)
        )
        return argument

    def stream(self, args=None, context=None):
        if args is None:
            args = CliArgs()
        elif isinstance(args, (tuple, list)):
            args = CliArgs(args)

        if context is None:
            context = {}

        # copy stack
        stack = list(self.stack.values())
        while stack:
            # try the first decorated function
            match = stack.pop(0)
            yield from match.stream(args)

        for arg in args:
            print(arg)


class GroupMismatch(Exception):
    pass


class ArgumentGroup:
    def __init__(self, fun, *arguments):
        self.fun = fun
        self.arguments = list(arguments)

    def add(self, argument):
        self.arguments.append(argument)

    def _consume(self, args):
        # test condiates
        arguments = list(self.arguments)
        while arguments:
            candidate = arguments.pop(0)
            match = candidate.matches(args)
            __import__('pdb').set_trace()      # XXX BREAKPOINT
            if match:
                yield match
            else:
                arg

        # restore args if failed

    def stream(self, args):
        yield from self._consume(args)


class Match:
    def __init__(self, argument, consumed_args, values):
        self.argument = argument
        self.consumed_args = consumed_args
        self.values = values


class Argument:
    def __init__(self, parse, *args, **kwargs):
        self.parse = parse
        self.args = args
        self.kwargs = kwargs

    def __call__(self, fun):
        try:
            self.parse.stack[fun].add(self)
        except KeyError:
            self.parse.stack[fun] = ArgumentGroup(fun, self)
        return fun

    def matches(self, args, context=None):
        if context is None:
            context = {}
        return None


class Option(Argument):
    def __init__(self, parse, name, *flags):
        super().__init__(parse)
        self.name = name
        self.flags = flags

    def matches(self, args, context=None):
        __import__('pdb').set_trace()      # XXX BREAKPOINT
        if len(args) > 1 and args[0] in self.flags:
            match = Match(self, args[:2], {self.name: args[1]})
            args = args[2:]
            return match
        return None





parse = Parse()


@parse.argument(Option, 'config', '-c', '--config')
def foo(config):
    print("foo")



def main():
    list(parse.stream())



import pytest


def test_config_source_schematize(mocker):
    from buvar import config
    import typing
    import attr

    @attr.s(auto_attribs=True)
    class FooConfig:
        bar: str = "default"
        foobar: float = 9.87
        baz: bool = config.bool_var(default=False)

    @attr.s(auto_attribs=True)
    class BarConfig:
        bim: float
        foo: FooConfig = FooConfig()

    @attr.s(auto_attribs=True, kw_only=True)
    class BimConfig:
        bar: BarConfig
        bam: bool = config.bool_var()
        bum: int = config.var(123)
        lst: typing.List = config.var(list)

    sources = [
        {"bar": {"bim": "123.4", "foo": {"bar": "1.23", "baz": "true"}}},
        {"foo": {"bar": "value", "foobar": 123.5, "baz": True}},
        {
            "bar": {"bim": "123.4"},
            "bim": {"bar": {"bim": 1.23}, "bam": "on", "lst": [1, 2, 3]},
        },
    ]

    mocker.patch(
        "os.environ",
        {
            "PREFIX_BAR_BIM": "0",
            "PREFIX_BAR_FOO_FOOBAR": "7.77",
            "PREFIX_BAR_FOO_BAZ": "false",
        },
    )

    cfg = config.ConfigSource(*sources, env_prefix="PREFIX")
    bar = cfg.load(BarConfig, "bar")
    foo = cfg.load(FooConfig, "foo")
    bim = cfg.load(BimConfig, "bim")

    assert (bar, foo, bim) == (
        BarConfig(bim=0.0, foo=FooConfig(bar="1.23", foobar=7.77, baz=False)),
        FooConfig(bar="value", foobar=123.5, baz=True),
        BimConfig(bar=BarConfig(bim=1.23), bam=True, lst=[1, 2, 3]),
    )


def test_load_general_config():
    import attr
    import typing
    from buvar import config

    sources = [{"foo": "bar", "group": {"some": "value"}}]
    cfg = config.ConfigSource(*sources, env_prefix="PREFIX")

    @attr.s(auto_attribs=True)
    class GeneralVars:
        foo: str
        baz: typing.Optional[float] = None

    general = cfg.load(GeneralVars)
    assert general == GeneralVars("bar", None)


def test_config_missing():
    import attr
    from buvar import config

    source = {"foo": {}}

    @attr.s(auto_attribs=True)
    class FooConfig:
        bar: str = config.var()

    cfg = config.ConfigSource(source)
    with pytest.raises(ValueError):
        cfg.load(FooConfig, "foo")


def test_generate_toml_help():
    import typing
    import attr
    from buvar import config

    @attr.s(auto_attribs=True)
    class FooConfig:
        """FooConfig.

        bim bam
        """

        string_val: str = config.var(help="string")
        float_val: float = config.var(9.87, help="float")
        bool_val: bool = config.bool_var(help="bool")
        int_val: int = config.var(help="int")
        list_val: typing.List = config.var(list, help="list")

    @attr.s(auto_attribs=True)
    class BarConfig:
        """BarConfig.

        bla bla
        bli bli
        """

        bim: float
        foo: FooConfig = config.var(help="foo")

    env_vars = list(config.traverse_attrs(BarConfig))

    assert [path for path, _ in env_vars] == [
        ("bim",),
        ("foo", "string_val"),
        ("foo", "float_val"),
        ("foo", "bool_val"),
        ("foo", "int_val"),
        ("foo", "list_val"),
    ]

    help = config.generate_toml_help(BarConfig, env_prefix="PREFIX")
    assert (
        help.as_string()
        == """# BarConfig.
# 
# bla bla
# bli bli

# bim =


[foo]
# FooConfig.
# 
# bim bam

# string
# string_val =

# float
float_val = 9.87

# bool
# bool_val =

# int
# int_val =

# list
list_val = []

"""
    )  # noqa: W291


def test_nested_attrs_typing():
    import typing
    import attr
    from buvar import config

    @attr.s(auto_attribs=True)
    class Baz:
        baz: str = "foobar"

    @attr.s(auto_attribs=True)
    class Bar:
        baz: Baz

    @attr.s(auto_attribs=True)
    class Foo:
        bars: typing.List[Bar] = []

    source = config.ConfigSource(
        {"foo": {"bars": [{"baz": {"baz": "something else"}}]}}, env_prefix="TEST"
    )

    foo = source.load(Foo, "foo")
    assert foo == Foo(bars=[Bar(baz=Baz(baz="something else"))])



import pytest


def test_run(event_loop):
    from buvar import plugin, components

    cmps = components.Components()

    plugin.run("tests.foo_plugin", components=cmps, loop=event_loop)

    assert cmps.get("foo_plugin") == {"foo": "foo"}


def test_run_relative_out_of_packages(event_loop):
    from buvar import plugin

    with pytest.raises(RuntimeError):
        plugin.run("tests.baz_plugin", loop=event_loop)


def test_run_get_event_loop(event_loop, mocker):
    from buvar import plugin

    mocker.patch.object(plugin.asyncio, "get_event_loop").return_value = event_loop

    async def test_plugin():
        pass

    plugin.run(test_plugin)


def test_run_load_twice(event_loop):
    from buvar import plugin

    loaded = {}

    async def test_plugin(include):
        assert not loaded
        loaded[True] = True
        await include(test_plugin)

    plugin.run(test_plugin)


def test_run_iterable(event_loop):
    from buvar import plugin

    state = {}

    async def iter_plugin():
        async def a():
            state["a"] = True

        async def b():
            state["b"] = True

        return [a(), b()]

    plugin.run(iter_plugin, loop=event_loop)
    assert state == {"a": True, "b": True}


def test_run_server(event_loop, caplog):
    import asyncio
    from buvar import plugin, components, context

    cmps = components.Components()

    async def stop_server_on_start():
        fut_server = context.get(asyncio.Future, name="server")
        evt_server_started = context.get(asyncio.Event, name="server_started")
        await evt_server_started.wait()
        fut_server.cancel()
        await fut_server
        assert "Server stopped" in caplog.text

    async def test_plugin(include):
        await include("tests.server_plugin")
        yield stop_server_on_start()

    plugin.run(test_plugin, components=cmps, loop=event_loop)
    assert "Server started" in caplog.text


def test_plugin_error(event_loop):
    from buvar import plugin

    async def broken_plugin():
        raise Exception("Plugin is broken")

    with pytest.raises(Exception) as e:
        plugin.run(broken_plugin, loop=event_loop)
        assert e.error.args == ("Plugin is broken",)


def test_cancel_main_task(event_loop):
    import asyncio
    from buvar import plugin, components

    cmps = components.Components()
    state = {}

    async def server_plugin():
        evt_cancel_main_task = cmps.get(plugin.CancelMainTask)

        async def server():
            async def teardown():
                state["teardown"] = True

            try:
                state["server"] = True
                evt_cancel_main_task.set()
                await asyncio.Future()
            except asyncio.CancelledError:
                assert True
            else:
                assert False

            yield teardown()

        yield server()

        evt_plugins_loaded = cmps.get(plugin.PluginsLoaded)

        async def wait_for_plugins_loaded():
            await evt_plugins_loaded.wait()
            # shutdown
            evt_cancel_main_task.set()

        # yield wait_for_plugins_loaded()

    plugin.run(server_plugin, components=cmps, loop=event_loop)
    assert state == {"teardown": True, "server": True}


def test_resolve_dotted_name():
    from buvar import plugin

    fun = plugin.resolve_dotted_name("dotted:name")

    assert fun() == "foobar"


def test_resolve_dotted_name_no_module():
    from buvar import plugin

    with pytest.raises(ValueError):
        plugin.resolve_dotted_name("nomodule:name")


def test_resolve_dotted_name_wrong_name():
    from buvar import plugin

    with pytest.raises(ValueError):
        plugin.resolve_dotted_name("nomodule:na:me")


def test_resolve_plugin_not_async(event_loop):
    from buvar import plugin

    with pytest.raises(ValueError):
        plugin.resolve_plugin(lambda: None)


def test_teardown(event_loop):
    from buvar import plugin, components

    cmps = components.Components()

    state = {}

    async def test_plugin():
        state["plugin"] = True

        async def task():
            state["task"] = True

            async def teardown():
                state["teardown"] = True

            return teardown()

        async def task2():
            async def teardown2():
                state["teardown2"] = True

            yield teardown2()

        async def task3():
            async def teardown3():
                state["teardown3"] = True

            return [teardown3()]

        yield task()
        yield task2()
        yield task3()

    plugin.run(test_plugin, components=cmps, loop=event_loop)

    assert state == {
        "plugin": True,
        "task": True,
        "teardown": True,
        "teardown2": True,
        "teardown3": True,
    }



import collections
import functools
import inspect
import itertools
import typing

import attr
import typing_inspect

from .. import util


class ResolveError(Exception):
    pass


def _extract_optional_type(hint):
    if typing_inspect.is_optional_type(hint):
        hint, _ = typing_inspect.get_args(hint)
        return hint
    return hint


@attr.s(auto_attribs=True)
class Adapter:
    func: typing.Callable
    spec: inspect.FullArgSpec
    frame: typing.Any
    owner: typing.Optional[typing.Type]
    name: typing.Optional[str]
    implements: typing.Type
    defaults: typing.Dict[str, typing.Any]

    @property
    def hints(self):
        if inspect.isfunction(self.func):
            if self.owner is not None:
                frame = self.frame.f_back.f_back
            else:
                frame = self.frame.f_back
        else:  # if inspect.isclass(self.func):
            frame = self.frame.f_back.f_back

        f_locals = dict(frame.f_locals)
        if isinstance(self.implements, str) and self.implements not in f_locals:
            f_locals[self.implements] = self.owner
        hints = typing.get_type_hints(
            self.func.__init__ if inspect.isclass(self.func) else self.func,
            frame.f_globals,
            f_locals,
        )
        return hints

    @util.reify
    def annotations(self):
        hints = self.hints
        annotations = {
            arg: _extract_optional_type(hints[arg])
            for arg in itertools.chain(self.spec.args, self.spec.kwonlyargs)
        }

        return annotations

    async def create(self, *args, **kwargs):
        """Create the target instance."""
        if self.owner is not None:
            func = functools.partial(self.func, self.owner)
        else:
            func = self.func
        if inspect.iscoroutinefunction(self.func):
            adapted = await func(*args, **kwargs)
        else:
            adapted = func(*args, **kwargs)
        return adapted


adapters: typing.Dict[type, typing.List[Adapter]] = collections.defaultdict(list)



import aiohttp.web
import attr
import structlog

from buvar import config, context


@attr.s(auto_attribs=True)
class AioHttpConfig:
    host: str = "0.0.0.0"
    port: int = 8080


class AccessLogger(aiohttp.abc.AbstractAccessLogger):  # noqa: R0903
    def log(self, request, response, time):  # noqa: R0201
        log = structlog.get_logger()
        log.info(
            "Access",
            remote=request.remote,
            method=request.method,
            path=request.path,
            time=time,
            status=response.status,
        )


async def plugin():
    config_source = context.get(config.ConfigSource)
    aiohttp_config = config_source.load(AioHttpConfig, "aiohttp")
    context.add(aiohttp_config)

    aiohttp_app = aiohttp.web.Application(
        middlewares=[aiohttp.web.normalize_path_middleware()]
    )
    context.add(aiohttp_app)

    yield aiohttp.web._run_app(  # noqa: W0212
        aiohttp_app, host=aiohttp_config.host, port=aiohttp_config.port, print=None
    )



from buvar import context


async def foo():
    ctx = context.find(str)
    context.add(ctx, "foo_plugin")


async def plugin(include):
    await include(".bar_plugin:plugin_bar")
    context.add("foo", name="foo")

    return foo()



from __future__ import annotations

import pytest


@pytest.fixture(autouse=True)
def mock_adapters(mocker):
    import collections

    mocker.patch("buvar.di.adapter.adapters", collections.defaultdict(list))


@pytest.fixture
def event_loop(event_loop):
    from buvar import context, components

    cmps = components.Components()
    context.set_task_factory(cmps, loop=event_loop)
    return event_loop


@pytest.fixture(params=["resolve", "c_resolve"])
def nject(request, mocker):
    if request.param == "resolve":
        from buvar.di import resolve

        mocker.patch("buvar.di.nject", resolve.nject)
        yield resolve.nject
    else:
        try:
            from buvar.di import c_resolve as resolve

            mocker.patch("buvar.di.nject", resolve.nject)
            yield resolve.nject  # noqa: I1101
        except ImportError:
            pytest.skip(f"C extension {request.param} not available.")
            return


@pytest.mark.asyncio
async def test_di_nject(nject):
    from buvar import context, di

    class Foo(dict):
        def __init__(self, name=None):
            super().__init__(name=name, foo=True)

    @di.register
    class Bar(dict):
        def __init__(self, bar: Foo):
            super().__init__(foo=bar, bar=True)

    class Baz(dict):
        def __init__(self, **kwargs):
            super().__init__(baz=True, **kwargs)

    class Bim:
        pass

    class Bum(dict):
        def __init__(self):
            super().__init__(bum=True)

    @di.register
    async def baz_adapter(
        bar: Bar, bam: Foo = 1, bim: Bim = "default", *, bum: Bum, foo: Foo = None
    ) -> Baz:
        return Baz(foo=foo, bam=bam, bim=bim, bar=bar, bum=bum)

    context.add(Foo())
    context.add(Foo(name="bar"), name="bar")

    baz, bum = await nject(Baz, Bum, bum=Bum())
    assert isinstance(baz, Baz)
    assert baz == {
        "baz": True,
        "bam": {"foo": True, "name": None},
        "foo": {"foo": True, "name": None},
        "bim": "default",
        "bum": {"bum": True},
        "bar": {"bar": True, "foo": {"foo": True, "name": "bar"}},
    }
    assert bum == {"bum": True}


@pytest.mark.asyncio
async def test_di_register():
    from buvar import di

    class Bar:
        @di.register
        def adapt(cls) -> Bar:
            assert cls == Bar
            return cls()

    @di.register
    class Foo:
        def __init__(self, bar: Bar):
            self.bar = bar

        @di.register
        def adapt(cls, bar: Bar) -> Foo:
            return cls(bar)

    @di.register
    def adapt(bar: Bar) -> Foo:
        return Foo(bar)

    bar = await di.nject(Bar)
    foo = await di.nject(Foo)
    assert set(di.adapter.adapters) == {"Foo", Foo, "Bar", Bar}


@pytest.mark.asyncio
async def test_readme():
    from buvar import di

    class Bar:
        pass

    class Foo:
        def __init__(self, bar: Bar = None):
            self.bar = bar

        @di.register
        async def adapt(cls, baz: str) -> Foo:
            return Foo()

    @di.register
    async def adapt(bar: Bar) -> Foo:
        foo = Foo(bar)
        return foo

    foo = await di.nject(Foo, baz="baz")
    assert foo.bar is None

    bar = Bar()
    foo = await di.nject(Foo, bar=bar)
    assert foo.bar is bar



"""Very simple plugin architecture for asyncio.

    >>> components = Components()
    >>> loop = asyncio.get_event_loop()
    >>> async def plugin(load):
    ...     async def some_task():
    ...         async def teardown():
    ...             pass
    ...         yield teardown()
    ...     yield some_task()


    >>> stages = Staging(components=components, loop=loop)
    >>> load = next(stages)
    >>> load(plugin)

    >>> # wait for main task to finish
    >>> teardown_tasks = next(stages)

    >>> # finsh
    >>> next(stages)


    >>> stages = Staging(plugin, components=components, loop=loop)
    >>> for stage in stages:
    ...     pass
"""
import contextlib
import asyncio
import collections
import collections.abc
import importlib
import inspect
import itertools
import sys

import structlog

from . import context
from .components import Components

PLUGIN_FUNCTION_NAME = "plugin"


class Tasks(collections.OrderedDict):
    pass


class MainTaskFinished(asyncio.Event):
    pass


class PluginsLoaded(asyncio.Event):
    pass


class CancelMainTask(asyncio.Event):
    pass


class Bootstrap:
    """Maintain the plugin loading state."""

    def __init__(self, *, components=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.components = components if components is not None else Components()

        # set task factory to serve our components context
        context.set_task_factory(self.components, loop=self.loop)

        # a collection of all tasks to run
        self.collected_tasks = self.components.add(Tasks())
        self.evt_plugins_loaded = self.components.add(PluginsLoaded(loop=self.loop))

        # schedule the main task, which waits until all plugins are loaded
        self.fut_main_task = asyncio.ensure_future(self._run_tasks(), loop=self.loop)
        self.fut_cancel_main_task = asyncio.ensure_future(
            self._cancel_main_task(), loop=self.loop
        )

        self.evt_cancel_main_task = self.components.add(CancelMainTask(loop=self.loop))
        self.evt_main_task_finished = self.components.add(
            MainTaskFinished(loop=self.loop)
        )

    async def _cancel_main_task(self):
        try:
            await self.evt_cancel_main_task.wait()
        except asyncio.CancelledError:
            pass
        else:
            self.fut_main_task.cancel()
            # await self.fut_main_task

    async def include(self, plugin):
        """Hook the plugin from another plugin.

        The plugin may return an awaitable, an iterable of awaitable or an
        asyncgenerator of awaitables.

        :param plugin: the plugin to load
        :type plugin: a callable which may have one argument, receiving the
        include callback
        """
        plugin = resolve_plugin(plugin)
        if plugin not in self.collected_tasks:
            # mark plugin as loaded for recursive circular stuff
            self.collected_tasks[plugin] = []
            # load plugin
            spec = inspect.getfullargspec(plugin)
            result = plugin(self.include) if len(spec.args) == 1 else plugin()
            callables = [fun async for fun in generate_async_result(result)]
            self.collected_tasks[plugin].extend(callables)

    def load(self, plugin):
        """Load a plugin."""
        try:
            self.loop.run_until_complete(self.include(plugin))
        except Exception as ex:  # noqa: E722
            # if a plugin raises an error, we cancel the scheduled main task
            import traceback

            sl = structlog.get_logger()
            sl.error(
                "Error while loading plugins",
                exception=ex,
                traceback=traceback.format_exc(),
            )
            # cancel prepared main task
            self.fut_main_task.cancel()
            self.loop.run_until_complete(self.fut_main_task)
            raise

    @contextlib.contextmanager
    def run(self):
        """Run all collected tasks on context exit."""
        try:
            yield self.fut_main_task
        finally:
            self.evt_plugins_loaded.set()

    @property
    def collected_tasks_list(self):
        collected_tasks = list(itertools.chain(*self.collected_tasks.values()))
        return collected_tasks

    async def _run_tasks(self):
        # we wait for plugin loading completed
        try:
            await self.evt_plugins_loaded.wait()
        except asyncio.CancelledError:
            # stop waiting to cancel
            self.fut_cancel_main_task.cancel()
            await self.fut_cancel_main_task
            # we were in plugin loading stage
            # we cancel unawaited tasks
            for task in self.collected_tasks_list:
                if inspect.iscoroutine(task):
                    task.close()
                else:
                    task.cancel()

        # run all tasks together
        # we send them into background and gather their results
        running_tasks = {
            asyncio.ensure_future(
                async_gen_to_list(task) if inspect.isasyncgen(task) else task
            )
            for task in self.collected_tasks_list
        }
        try:
            # we run until all tasks have completed
            running_tasks_results = await asyncio.shield(
                asyncio.gather(*running_tasks, return_exceptions=True, loop=self.loop)
            )
        except asyncio.CancelledError:
            # stop all tasks
            for task in running_tasks:
                if not task.done():
                    task.cancel()
            running_tasks_results = await asyncio.gather(
                *running_tasks, return_exceptions=True, loop=self.loop
            )
        self.evt_main_task_finished.set()
        # retrieve results in any way
        teardown_tasks = list(flatten_tasks(running_tasks_results))
        self.fut_cancel_main_task.cancel()
        await self.fut_cancel_main_task
        return teardown_tasks


class Staging:
    """Generate all stages to run an application."""

    def __init__(self, *plugins, components=None, loop=None):
        self.components = components if components is not None else Components()
        self.loop = loop or asyncio.get_event_loop()

        # stage 1: bootstrap plugins
        self.bootstrap = self.components.add(
            Bootstrap(components=components, loop=loop)
        )
        self.plugins = set(plugins)

        def iterator():
            with self.bootstrap.run() as fut_main_task:
                for plugin in self.plugins:
                    self.bootstrap.load(plugin)
                # we yield load to additionally load arbitrary stuff, after standard
                # plugin loading
                yield self.bootstrap.load

            # stage 2: run main task
            teardown_tasks = self.loop.run_until_complete(fut_main_task)
            yield teardown_tasks

            # stage 3: teardown
            self.loop.run_until_complete(
                asyncio.gather(*teardown_tasks, loop=self.loop)
            )
            yield None

        self.iterator = iterator()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def cancel(self):
        self.components.get(CancelMainTask).set()


def run(*plugins, components=None, loop=None):
    """Start the asyncio process by boostrapping the root plugin.

    We have a three phase setup:

    1. run plugin hooks, to register arbitrary stuff and mainly gather a set of
    asyncio tasks in the second phase

    2. run all registered tasks together until complete

    3. run returned tasks for teardown
    """
    list(Staging(*plugins, components=components, loop=loop))


def resolve_plugin(plugin, function_name=PLUGIN_FUNCTION_NAME):
    plugin = resolve_dotted_name(plugin)
    if inspect.ismodule(plugin):
        # apply default name
        plugin = getattr(plugin, function_name)

    if not (inspect.iscoroutinefunction(plugin) or inspect.isasyncgenfunction(plugin)):
        raise ValueError(f"{plugin} must a coroutine or an async generator.")

    return plugin


def resolve_dotted_name(name):
    """Use pkg_resources style dotted name to resolve a name."""
    # skip resolving for module and coroutine
    if inspect.ismodule(name) or inspect.isroutine(name):
        return name

    # relative import
    if name.startswith("."):
        target_name = name.split(".")[1:]
        frame = inspect.currentframe()
        while frame.f_globals["__name__"].startswith(__package__):
            frame = frame.f_back

        caller_name = frame.f_globals["__package__"].split(".")
        try:
            while not target_name[0]:
                caller_name.pop()
                target_name.pop(0)
            name = ".".join(itertools.chain(caller_name, target_name))
        except IndexError:
            raise RuntimeError("Relative name gets out of parent packages!", name)
    part = ":"
    module_name, _, attr_name = name.partition(part)
    if part in attr_name:
        raise ValueError(f"Invalid name: {name}", name)

    if module_name in sys.modules:
        resolved = sys.modules[module_name]
    else:
        spec = importlib.util.find_spec(module_name)
        if spec is None:
            raise ValueError(f"Invalid module: {module_name}", module_name)
        resolved = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(resolved)
        sys.modules[resolved.__name__] = resolved

    if attr_name:
        resolved = getattr(resolved, attr_name)

    return resolved


async def generate_async_result(fun_result):
    """Create an async generator out of any result."""
    if inspect.isasyncgen(fun_result):
        async for item in fun_result:
            yield item
    else:
        result = await fun_result

        # store tasks
        if isinstance(result, collections.abc.Iterable):
            for item in result:
                yield item
        elif result is not None:
            yield result


async def async_gen_to_list(asyncgen):
    lst = [item async for item in asyncgen]
    return lst


def flatten_tasks(result_list):
    for item in result_list:
        if inspect.isawaitable(item):
            yield item
        elif isinstance(item, collections.abc.Iterable):
            for sub_item in item:
                if inspect.isawaitable(sub_item):
                    yield sub_item



"""A simple component architecture.

must have:
- add components by their type and a name
- add components by someother discriminator
- lookup components by their type/discriminator and name
- find components just by their type


nice to have:
- adaption resp. dependency injection of components
"""
import collections
import inspect
import structlog


sl = structlog.get_logger()
missing = object()


class ComponentLookupError(Exception):
    pass


class Components(collections.ChainMap):

    """A component registry holds certain items identified by a
    namespace discriminator."""

    def add(self, item, namespace=None, *, name=None):
        """Register `item` and optionally name it."""
        if inspect.isclass(item):
            raise ValueError("A component should be an instance.")
        if namespace is None:
            namespace = type(item)

        if namespace in self:
            self[namespace][name] = item
        else:
            self[namespace] = {name: item}
        return item

    def get(self, namespace, *, name=None, default=missing):
        try:
            item = self.find(namespace)[name]
        except KeyError:
            if default is missing:
                raise ComponentLookupError(
                    "Component not found", namespace, name, default
                )
            return default
        return item

    def find(self, namespace):
        """Try to match the discriminator.

        If more than one component matches the discriminator, we return the
        closest one.
        """
        if inspect.isclass(namespace):
            items = None
            for key in self.keys():
                if inspect.isclass(key) and issubclass(key, namespace):
                    if items is None or issubclass(items, key):
                        items = key

        else:
            items = namespace

        return self[items]



"""Dependency injection."""
import functools
import inspect
import itertools

from . import adapter


class register:
    """Register a class, function or method to adapt arguments into something else.

    >>> class Bar:
    ...     pass

    >>> @register
    ... class Foo:
    ...     def __init__(self, bar: Bar):
    ...         self.bar = bar

    ...     @register
    ...     def adapt(cls, bar: Bar) -> Foo:
    ...         return cls(bar)

    >>> @register
    ... def adapt(bar: Bar) -> Foo:
    ...     return Foo(bar)

    """

    def __new__(cls, func):
        instance = super().__new__(cls)
        if inspect.isclass(func):
            instance.__init__(func)
            return func
        return instance

    def __init__(self, func):
        currentframe = inspect.currentframe()
        spec = inspect.getfullargspec(func)
        if inspect.isclass(func):
            # we remove self from class specs
            spec.args.pop(0)

        implements = func if inspect.isclass(func) else spec.annotations["return"]

        # all none defaults must be annotated
        defaults = dict(
            itertools.chain(
                zip(reversed(spec.args or []), reversed(spec.defaults or [])),
                (spec.kwonlydefaults or {}).items(),
            )
        )

        self.adapter = adapter.Adapter(
            func=func,
            spec=spec,
            frame=currentframe,
            # globals=frame.f_globals,
            # locals=frame.f_locals,
            owner=None,
            name=None,
            defaults=defaults,
            implements=implements,
        )
        adapter.adapters[implements].append(self.adapter)

    def __set_name__(self, owner, name):
        self.adapter.owner = owner
        self.adapter.name = name
        # remove cls from args
        self.adapter.spec.args.pop(0)

        try:
            hints = self.adapter.hints
            self.adapter.implements = hints["return"]
            adapter.adapters[hints["return"]].append(self.adapter)
        except NameError:
            pass

    def __get__(self, instance, owner=None):
        if owner is None:
            owner = type(instance)
        return functools.partial(self.adapter.func, owner)

    def __call__(self, *args, **kwargs):
        return self.adapter.func(*args, **kwargs)


try:
    # gains over 100% speed up
    from .c_resolve import nject
except ImportError:
    from .resolve import nject  # noqa: W0611



"""Provide a component registry as task context."""
import asyncio
import contextlib
import sys

from . import components

__all__ = ("get", "add", "find")


PY37 = sys.version_info >= (3, 7)

if PY37:

    def asyncio_current_task(loop=None):
        """Return the current task or None."""
        try:
            return asyncio.current_task(loop)
        except RuntimeError:
            # simulate old behaviour
            return None


else:
    asyncio_current_task = asyncio.Task.current_task


class TaskFactory:
    def __init__(self, *, root_context=None, default=None, parent_factory=None):
        self.default = default or self.default_components_context
        self.factory = (
            parent_factory if parent_factory is not None else asyncio.tasks.Task
        )
        self.root_context = (
            root_context if root_context is not None else components.Components()
        )

    def default_components_context(self, parent_context=None):
        """Context is always the same as the parents one."""
        if parent_context:
            return parent_context

        return self.root_context

    def __call__(self, loop, coro):
        task = self.factory(loop=loop, coro=coro)
        context = current_context(loop=loop)

        setattr(task, "context", self.default(context))
        return task


def set_task_factory(components, *, loop=None):  # noqa: W0621
    if loop is None:
        loop = asyncio.get_event_loop()

    task_factory = TaskFactory(
        root_context=components, parent_factory=loop.get_task_factory()
    )

    loop.set_task_factory(task_factory)


def current_context(loop=None):
    current_task = asyncio_current_task(loop=loop)
    context = getattr(current_task, "context", None)
    return context


def add(*args, **kwargs):
    context = current_context()
    return context.add(*args, **kwargs)


def get(*args, **kwargs):
    context = current_context()
    return context.get(*args, **kwargs)


def find(*args, **kwargs):
    context = current_context()
    return context.find(*args, **kwargs)


def push():
    current_task = asyncio_current_task()
    parent_context = getattr(current_task, "context", None)
    assert parent_context is not None, "There must be a context to push from."

    context = parent_context.new_child()
    setattr(current_task, "context", context)


def pop():
    current_task = asyncio_current_task()
    parent_context = getattr(current_task, "context", None)
    assert parent_context is not None, "There must be a context to pop from."

    context = parent_context.parents
    setattr(current_task, "context", context)


@contextlib.contextmanager
def child():
    push()
    yield
    pop()



from . import adapter
from .. import components

missing = object()


async def nject(*targets, **dependencies):
    """Resolve all dependencies and return the created component."""
    from buvar import context

    # create components
    cmps = components.Components()

    # add default unnamed dependencies
    # every non-default argument of the same type gets its value
    # XXX is this good?
    for name, dep in dependencies.items():
        cmps.add(dep)

    # add current context
    cmps = cmps.new_child(context.current_context())

    # add default named dependencies
    cmps = cmps.new_child()
    for name, dep in dependencies.items():
        cmps.add(dep, name=name)

    # find the proper components to instantiate that class
    injected = [
        await resolve_adapter(cmps, target, name=name)
        for name, target in ((None, target) for target in targets)
    ]
    if len(targets) == 1:
        return injected[0]
    return injected


def find_string_target_adapters(target):
    name = target.__name__

    # search for string and match
    string_adapters = adapter.adapters.get(name)
    # find target name in string adapter types and find target in adapter
    if string_adapters:
        for adptr in string_adapters:
            if adptr.hints["return"] is target:
                yield adptr


async def resolve_adapter(cmps, target, *, name=None, default=missing):
    # find in components
    try:
        component = _get_name_or_default(cmps, target, name)
        return component
    except components.ComponentLookupError:
        pass

    # try to adapt
    possible_adapters = (adapter.adapters.get(target) or []) + list(
        find_string_target_adapters(target)
    )

    resolve_errors = []
    if possible_adapters is None:
        if default is missing:
            raise adapter.ResolveError(
                "No possible adapter found", target, resolve_errors
            )
        return default

    for adptr in possible_adapters:
        try:
            adapter_args = {
                name: await resolve_adapter(
                    cmps,
                    dependency_target,
                    name=name,
                    default=adptr.defaults.get(name, missing),
                )
                for name, dependency_target in adptr.annotations.items()
            }
        except adapter.ResolveError as ex:
            # try next adapter
            resolve_errors.append(ex)
        else:
            component = await adptr.create(**adapter_args)
            # we do not use the name
            cmps.add(component)
            return component
    if default is missing:
        raise adapter.ResolveError(
            "No adapter dependencies found", target, resolve_errors
        )
    return default


def _get_name_or_default(cmps, target, name=None):
    # find in components
    if name is not None:
        try:
            component = cmps.get(target, name=name)
            return component
        except components.ComponentLookupError:
            pass

    component = cmps.get(target, name=None)
    return component



import asyncio

from buvar import context


async def plugin(load):
    import structlog

    logger = structlog.get_logger()

    fut = asyncio.Future()
    context.add(fut, name="server")
    evt_server_started = asyncio.Event()
    context.add(evt_server_started, name="server_started")

    async def server(fut):
        logger.info("Server started")
        evt_server_started.set()
        await fut
        logger.info("Server stopped")

    task = server(fut)
    return task



import pytest


def test_lookup_mro():
    from buvar import components

    class Foo:
        pass

    class Bar(Foo):
        pass

    class Baz(Bar):
        pass

    c = components.Components()

    foo = Foo()
    baz = Baz()
    c.add(foo)
    c.add(baz)

    assert c.get(Foo) == foo
    assert c.get(Bar) == baz
    assert c.get(Baz) == baz


def test_lookup_basic():
    from buvar import components
    c = components.Components()

    c.add('foo', 'bar')

    assert c.get('bar') == 'foo'


def test_no_instance():
    from buvar import components
    c = components.Components()

    class Foo:
        pass

    with pytest.raises(ValueError):
        c.add(Foo)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['buvar', 'buvar.di', 'buvar.plugins']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0',
 'cattrs>=0.9.0,<0.10.0',
 'multidict>=4.5,<5.0',
 'orjson>=2.0,<3.0',
 'structlog>=19.1,<20.0',
 'toml>=0.10,<0.11',
 'tomlkit>=0.5.3,<0.6.0',
 'typing_inspect>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['buvar = buvar.cli:main']}

setup_kwargs = {
    'name': 'buvar',
    'version': '0.10.0',
    'description': 'General purpose asyncio service loader and dependency injector',
    'long_description': 'Búvár\n=====\n\na plugin mechanic\n-----------------\n\nI want to have a plugin mechanic similar to `Pyramid`_\\\'s aproach. It should\nprovide means to spawn arbitrary tasks to run, where every lifecycle stage\nshould be yielded to the developer control.\n\nYou bootstrap like following:\n\n.. code-block:: python\n\n    from buvar import plugin, components\n\n    plugin.run("some.module.with.plugin.function")\n\n\n.. code-block:: python\n\n   # some.module.with.plugin.function\n   from buvar import context\n\n\n   # you may omit include in arguments\n   async def plugin(include):\n      await include(\'.another.plugin\')\n\n      # create some long lasting components\n      my_component = context.add("some value")\n\n      async def task():\n         asyncio.sleep(1)\n\n      async def server():\n         await asyncio.Future()\n\n      # you may run simple tasks\n      yield task()\n\n      # you may run server tasks\n      yield server()\n\n\na components and dependency injection solution\n----------------------------------------------\n\nI want to have some utility to store some long lasting means of aid to my\nbusiness problem. I want a non-verbose lookup to those.\n\n.. code-block:: python\n\n   from buvar import di\n\n   class Bar:\n      pass\n\n   class Foo:\n      def __init__(self, bar: Bar = None):\n         self.bar = bar\n\n      @di.register\n      async def adapt(cls, baz: str) -> Foo:\n         return Foo()\n\n   @di.register\n   async def adapt(bar: Bar) -> Foo\n      foo = Foo(bar)\n      return foo\n\n\n   async def task():\n      foo = await di.nject(Foo, baz="baz")\n      assert foo.bar is None\n\n      bar = Bar()\n      foo = await di.nject(Foo, bar=bar)\n      assert foo.bar is bar\n\n\na config source\n---------------\n\nI want to have a config source, which automatically applies environment\nvariables to the defaults.\n\n`config.toml`\n\n.. code-block:: toml\n\n   log_level = "DEBUG"\n   show_warnings = "yes"\n\n   [foobar]\n   some = "value"\n\n\n.. code-block:: bash\n\n   export APP_FOOBAR_SOME=thing\n\n\n.. code-block:: python\n\n   import attr\n   import toml\n\n   from buvar import config\n\n   @attr.s(auto_attribs=True)\n   class GeneralConfig:\n       log_level: str = "INFO"\n       show_warnings: bool = config.bool_var(False)\n\n   @attr.s(auto_attribs=True)\n   class FoobarConfig:\n      some: str\n\n   source = config.ConfigSource(toml.load(\'config.toml\'), env_prefix="APP")\n\n   general_config = source.load(GeneralConfig)\n   assert general_config == GeneralConfig(log_level="DEBUG", show_warnings=True)\n\n   foobar_config = source.load(FoobarConfig, \'foobar\')\n   assert foobar_config.some == "thing"\n\n\n\na structlog\n-----------\n\nI want to have a nice and readable `structlog`_ in my terminal and a json log in\nproduction.\n\n.. code-block:: python\n\n   import sys\n\n   from buvar import log\n\n   log.setup_logging(sys.stdout.isatty(), general_config.log_level)\n\n\n.. _Pyramid: https://github.com/Pylons/pyramid\n.. _structlog: https://www.structlog.org/en/stable/\n',
    'author': 'Oliver Berger',
    'author_email': 'diefans@gmail.com',
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
